'''
Author: Ana Ferreira, <anafsferreira2@gmail.com>

Dataset: PhysioIntent

This files has some auxiliar function to help reading the data.

'''
import os
import datetime
import pandas as pd
import csv
import numpy as np
from brainflow.data_filter import DataFilter
import ast
import scipy.signal

def read_openBCI_data(brainflow, num_channels, file_path, openBCI_filename, timestamp_correction, sample_rate, new_acquisition):
    
    ''' 
    This function reads the openBCI data.

    Parameters
    ----------
    brainflow: bool
        If True the data file was acquired with Brainflow library. If False it was acquired with the GUI.
    num_channels: int
        If only used Cyton board: num_channels = 8
        If Daisy + Cyton: num_channels = 16
    file_path: string
        The path where the .csv files are. '/openBCI/
    openBCI_filename: string
        Name of the .csv file
    timestamp_correction: bool
        If False, reading raw data, no timestamp correction was yet done
        If True, reading data from timestamp_correction_openBCI/ folder
    sample_rate: int
        250 Hz if only cyton board
        125 Hz if daisy + cyton
    new_acquisition: bool
        True: if phase 2 acquisition

    Returns
    -------

    openBCI_X_df: dataframe
        n columns: ['index', 'fp1', 'fp2', 'fc3', 'fcz', 'fc4', 'c1','c3', 'cz', 'c2', 'c4', 'cp3', 'cp4', 'p3', 'p4', 'ecg', 'emg_extensor','acc1', 'acc2', etc]
    '''

    if brainflow:
        if not timestamp_correction:

            #read openBCI file
            openBCI_brainflow_data = DataFilter.read_file( file_path + + 'raw_data_openBCI/' + openBCI_filename)
            openBCI_brainflow_df = pd.DataFrame(np.transpose(openBCI_brainflow_data))
            if num_channels == 8: # CYTHON
                openBCI_brainflow_df.columns = ['index', 'ch1', 'ch2', 'ch3', 'ch4', 'ch5', 'ch6', 'ch7', 'ch8', 'acc1', 'acc2', 'acc3', 'Other1', 'Other2', 'Other3', 'Other4', 'Other5', 'Other6', 'Other7', 'A0', 'A1','A2','Time', 'Other']
            elif num_channels == 16: # CYTHON + DAISY
                if new_acquisition: #different channels
                    openBCI_brainflow_df.columns = ['index', 'fp1', 'fp2', 'fc3', 'fcz', 'fc4', 'c1','c3', 'cz', 'c2', 'c4', 'cp3', 'cp4', 'p3', 'p4', 'ecg', 'emg_extensor','acc1', 'acc2', 'acc3', 'Other1', 'Other2', 'Other3', 'Other4', 'Other5', 'Other6', 'Other7', 'A0', 'A1','A2','Time', 'Other']
                else:
                    openBCI_brainflow_df.columns = ['index', 'fp1', 'fp2', 'fcz', 'c3', 'cz', 'c4', 'cp3', 'cp4', 'p3', 'p4', 'ecg', 'ch12', 'emg_bicep', 'emg_tricep', 'emg_flexor', 'emg_extensor','acc1', 'acc2', 'acc3', 'Other1', 'Other2', 'Other3', 'Other4', 'Other5', 'Other6', 'Other7', 'A0', 'A1','A2','Time', 'Other']
            
            #convert timestamp
            timestamp = pd.to_datetime(openBCI_brainflow_df["Time"], unit = 's')

            ##add a timestamp columns 
            openBCI_brainflow_df['TimeStamp']=  timestamp + datetime.timedelta(hours=1) 

            # correction fo timestamp
            stop = len(openBCI_brainflow_df['TimeStamp'])/sample_rate-1/sample_rate
            step = 1/sample_rate
            sampleTime_openBCI = np.arange(0, round(stop, 3) + step, step)
            # print("Length dataframe: ", len(openBCI_brainflow_df['TimeStamp']))
            # print("Timestamps shape: ", sampleTime_openBCI.shape)
            # print("Total length of data: ", sampleTime_openBCI[-1], 's')

            timestamp_openBCI= np.asarray([])
            for i in range(len(openBCI_brainflow_df['TimeStamp'])):
                timestamp_openBCI = np.append(timestamp_openBCI, openBCI_brainflow_df['TimeStamp'][0] + datetime.timedelta(seconds= sampleTime_openBCI[i]))

            openBCI_brainflow_df['TimeStamp']= timestamp_openBCI 
            #create folder timestamp_correction
            save_path = file_path + '/timestamp_correction_openBCI'
            CHECK_FOLDER = os.path.isdir(save_path)

            # If folder doesn't exist, then create it.
            if not CHECK_FOLDER:
                os.makedirs(save_path)
                print("created folder : ", save_path)
            else:
                print(save_path, "folder already exists.")

            # save csv
            file_save_path = file_path + 'timestamp_correction_openBCI/' + openBCI_filename
            openBCI_brainflow_df.to_csv(file_save_path)

            return openBCI_brainflow_df

        else:
            openBCI_brainflow_file = file_path + 'timestamp_correction_openBCI/' + openBCI_filename
            if num_channels == 8:
                openBCI_brainflow_df = pd.read_csv(openBCI_brainflow_file, sep=",", 
                    usecols=['ch1', 'ch2', 'ch3', 'ch4', 'ch5', 'ch6', 'ch7', 'ch8', 'A1', 'Time', 'TimeStamp']) 
            elif num_channels == 16:

                if new_acquisition:
                                    openBCI_brainflow_df = pd.read_csv(openBCI_brainflow_file, sep=",", 
                    usecols=['fp1', 'fp2', 'fc3', 'fcz', 'fc4', 'c1','c3', 'cz', 'c2', 'c4', 'cp3', 'cp4','p3', 'p4','ecg','emg_extensor', 'A1', 'Time', 'TimeStamp']) 

                else:
                    openBCI_brainflow_df = pd.read_csv(openBCI_brainflow_file, sep=",", 
                        usecols=['fp1', 'fp2', 'fcz', 'c3', 'cz', 'c4', 'cp3', 'cp4', 'p3', 'p4', 'ecg', 'emg_bicep', 'emg_tricep', 'emg_flexor', 'emg_extensor', 'A1', 'Time', 'TimeStamp']) 
            #print(openBCI_brainflow_df.head(10)) 
            return openBCI_brainflow_df
    else: # GUI
            
            if not timestamp_correction:
                openBCI_GUI_file = file_path + openBCI_filename
                if num_channels == 8:
                    openBCI_GUI_df = pd.read_csv(openBCI_GUI_file, sep=",", header=7, index_col=False, names=['index', 'ch1', 'ch2', 'ch3', 'ch4', 'ch5', 'ch6', 'ch7', 'ch8', 'acc1', 'acc2', 'acc3', 'Other1', 'Other2', 'Other3', 'Other4', 'Other5', 'Other6', 'Other7', 'A0', 'A1','A2','Time', 'Other', 'TimeStamp'], 
                    usecols=['ch1', 'ch2', 'ch3', 'ch4', 'ch5', 'ch6', 'ch7', 'ch8', 'A1', 'Time', 'TimeStamp']) 
                elif num_channels == 16:
                    openBCI_GUI_df = pd.read_csv(openBCI_GUI_file, sep=",", header=7, index_col=False, names=  ['index', 'fp1', 'fp2', 'fcz', 'c3', 'cz', 'c4', 'cp3', 'cp4', 'p3', 'p4', 'ecg', 'ch12', 'emg_bicep', 'emg_tricep', 'emg_flexor', 'emg_extensor','acc1', 'acc2', 'acc3', 'Other1', 'Other2', 'Other3', 'Other4', 'Other5', 'Other6', 'Other7', 'A0', 'A1','A2','Time', 'Other', 'TimeStamp'],
                    usecols=['fp1', 'fp2', 'fcz', 'c3', 'cz', 'c4', 'cp3', 'cp4', 'p3', 'p4', 'ecg', 'emg_bicep', 'emg_tricep', 'emg_flexor', 'emg_extensor', 'A1', 'Time', 'TimeStamp'])
                
                # correction fo timestamp
                openBCI_GUI_df['TimeStamp'] = pd.to_datetime(openBCI_GUI_df['TimeStamp'])
                stop = len(openBCI_GUI_df['TimeStamp'])/sample_rate-1/sample_rate
                step = 1/sample_rate
                sampleTime_openBCI = np.arange(0, round(stop, 3) + step, step)
                print("Length dataframe: ", len(openBCI_GUI_df['TimeStamp']))
                print("Timestamps shape: ", sampleTime_openBCI.shape)
                print("Total length of data: ", sampleTime_openBCI[-1], 's')
                timestamp_openBCI= np.asarray([])
                for i in range(len(openBCI_GUI_df['TimeStamp'])):
                    timestamp_openBCI = np.append(timestamp_openBCI, openBCI_GUI_df['TimeStamp'][0] + datetime.timedelta(seconds= sampleTime_openBCI[i]))

                openBCI_GUI_df['TimeStamp']= timestamp_openBCI 
                # save csv
                file_save_path = file_path + 'timestamp_correction_openBCI/' + openBCI_filename
                openBCI_GUI_df.to_csv(file_save_path)

                return openBCI_GUI_df

            else:
                openBCI_GUI_file = file_path + '/timestamp_correction_openBCI/' + openBCI_filename
                if num_channels == 8:
                    openBCI_GUI_df = pd.read_csv(openBCI_GUI_file, sep=",", 
                        usecols=['ch1', 'ch2', 'ch3', 'ch4', 'ch5', 'ch6', 'ch7', 'ch8', 'A1', 'Time', 'TimeStamp']) 
                elif num_channels == 16:
                    openBCI_GUI_df = pd.read_csv(openBCI_GUI_file, sep=",", 
                        usecols=['fp1', 'fp2', 'fcz', 'c3', 'cz', 'c4', 'cp3', 'cp4', 'p3', 'p4', 'ecg', 'emg_bicep', 'emg_tricep', 'emg_flexor', 'emg_extensor', 'A1', 'Time', 'TimeStamp'])
                return openBCI_GUI_df

def read_openBCI_filter_data(openBCI_path, openBCI_filename, new_acquisition):

    ''' 
    This function reads the openBCI data filtered.

    Parameters
    ----------
    openBCI_path: string
        The path where the .csv files are. '/openBCI/
    openBCI_filename: string
        Name of the .csv file
    new_acquisition: bool
        True: if phase 2 acquisition

    Returns
    -------

    openBCI_df: dataframe
        n columns: ['index', 'fp1', 'fp2', 'fc3', 'fcz', 'fc4', 'c1','c3', 'cz', 'c2', 'c4', 'cp3', 'cp4', 'p3', 'p4', 'ecg', 'emg_extensor','acc1', 'acc2', etc]
    '''

    openBCI_brainflow_file = openBCI_path + 'filter_data_openBCI/' + openBCI_filename

    if new_acquisition:
        openBCI_df = pd.read_csv(openBCI_brainflow_file, sep=',', 
                        usecols=['fp1', 'fp2', 'fc3', 'fcz', 'fc4', 'c1','c3', 'cz', 'c2', 'c4', 'cp3', 'cp4', 'p3', 'p4', 'ecg', 'emg_extensor', 'A1', 'Time', 'TimeStamp'])
    else: 
        openBCI_df = pd.read_csv(openBCI_brainflow_file, sep=",", 
                        usecols=['fp1', 'fp2', 'fcz', 'c3', 'cz', 'c4', 'cp3', 'cp4', 'p3', 'p4', 'ecg', 'emg_bicep', 'emg_tricep', 'emg_flexor', 'emg_extensor', 'A1', 'Time', 'TimeStamp']) 

    return openBCI_df


def read_iHandU_data(upsample, iHandU_path_filename, subject, sample_rate_iHandU, new_acquisition):

    ''' 
    This function upsamples the iHandU data to a .csv if the upsample is False. Otherwise, reads the upsample file and returns the dataframe

    Parameters
    ----------
    upsample: bool
        If True, already has the upsample done
    iHandU_path_filename: string
        All path including the filename
    subject: string
        'S1','S2', etc
    sample_rate_iHandU: int
        Sample rate in which the iHandU was record
    new_acquisition: bool
        If True phase 2 data, no lost of data on the iHandU device

    Returns
    -------

    df_raw_iHandU_upsample: dataframe
        2 columns: [name; data]
    acc_x, acc_y, acc_z, gir_x, gir_y, gir_z: array
    '''
    
    # iHandU path
    if iHandU_path_filename.endswith(iHandU_path_filename.split('/')[-1]):
        iHandU_path = iHandU_path_filename[:-len(iHandU_path_filename.split('/')[-1])]

    if not upsample:
        file = open(iHandU_path + subject+'_iHandU_upsample.csv', 'w', newline='') 
        writer = csv.writer(file, quoting=csv.QUOTE_NONNUMERIC, delimiter=',')
        writer.writerow(['name', 'data'])

        #read raw iHandU
        column_names = [i for i in range(0, 20)]
        df_raw_iHandU = pd.read_csv(iHandU_path_filename,',', header = None, names = column_names)

        df_time_iHandU =  df_raw_iHandU.iloc[:,0].astype(str)+ '-'+df_raw_iHandU.iloc[:,1].astype(str) +'-'+  df_raw_iHandU.iloc[:,2].astype(str) +' ' + df_raw_iHandU.iloc[:,3].astype(str) +':'+ df_raw_iHandU.iloc[:,4].astype(str) +':'+ df_raw_iHandU.iloc[:,5].astype(str)     
        df_time_iHandU = pd.to_datetime(df_time_iHandU)
        df_time_iHandU = df_time_iHandU + pd.to_timedelta(df_raw_iHandU.iloc[:,6], unit='ms')

        df_raw_iHandU = pd.concat([df_raw_iHandU.loc[:,7:15], df_time_iHandU], axis = 1)
        df_raw_iHandU.columns = ['accel_x', 'accel_y', 'accel_z', 'gir_x', 'gir_y', 'gir_z', 'mag_x', 'mag_y', 'mag_z', 'timestamp' ]

        duration_iHandU = (df_raw_iHandU.timestamp.iat[-1]- df_raw_iHandU.timestamp[1]).total_seconds()

        #calculate sample rate
        print('Duration of iHandU data: ', df_raw_iHandU.timestamp.iat[-1]- df_raw_iHandU.timestamp[1])
        print('Number of samples: ', df_raw_iHandU.shape)
        new_sample_rate_iHandU= len(df_raw_iHandU.timestamp)/duration_iHandU
        print("New sample rate iHandU: ", new_sample_rate_iHandU)

        if new_acquisition: #not lost of data, use the new_sample_rate
            sample_rate_iHandU = new_sample_rate_iHandU
        print("Sample rate to be use: ", sample_rate_iHandU)
        #save upsample file
        for (columnName, columnData) in df_raw_iHandU.iteritems():
            upsample_ihandU=  list(signal_resample(np.array(columnData.values[1:]), sampling_rate= sample_rate_iHandU, desired_sampling_rate= 125, method ='fft'))
            writer.writerow([columnName, upsample_ihandU])
        file.close()


    df_raw_iHandU_upsample = pd.read_csv(iHandU_path + subject+'_iHandU_upsample.csv',',')

    acc_x = np.array(ast.literal_eval(df_raw_iHandU_upsample['data'][0]))
    acc_y = np.array(ast.literal_eval(df_raw_iHandU_upsample['data'][1]))
    acc_z = np.array(ast.literal_eval(df_raw_iHandU_upsample['data'][2]))
    gir_x = np.array(ast.literal_eval(df_raw_iHandU_upsample['data'][3]))
    gir_y = np.array(ast.literal_eval(df_raw_iHandU_upsample['data'][4]))
    gir_z = np.array(ast.literal_eval(df_raw_iHandU_upsample['data'][5]))
    
    #print('Length iHandU data: ', acc_x.shape)

    return df_raw_iHandU_upsample, acc_x, acc_y, acc_z, gir_x, gir_y, gir_z



def signal_resample(signal, desired_length=None, sampling_rate=None, desired_sampling_rate=None, method="interpolation"):

    '''
        This function thus the signal resample

        PARAMETERS:
        ------------
        signal: array
        desired_length: int (samples)
        sampling_rate: int (Hz)
        desired_sampling_rate: int (Hz)
        method: string
            'fft', 'poly', 'pandas', etc

        RETURN:
        -------
        resampled: array
            Signal with new sample rate
    '''
    if desired_length is None:
        desired_length = int(np.round(len(signal) * desired_sampling_rate / sampling_rate))

    # Sanity checks
    if len(signal) == desired_length:
        return signal

    # Resample
    if method.lower() == "fft":
        resampled = _resample_fft(signal, desired_length)
    elif method.lower() == "poly":
        resampled = _resample_poly(signal, desired_length)
    elif method.lower() == "numpy":
        resampled = _resample_numpy(signal, desired_length)
    elif method.lower() == "pandas":
        resampled = _resample_pandas(signal, desired_length)
    else:
        resampled = _resample_interpolation(signal, desired_length)

    return resampled

def _resample_numpy(signal, desired_length):
    resampled_signal = np.interp(
        np.linspace(0.0, 1.0, desired_length, endpoint=False),  # where to interpolate
        np.linspace(0.0, 1.0, len(signal), endpoint=False),  # known positions
        signal,  # known data points
    )
    return resampled_signal


def _resample_interpolation(signal, desired_length):
    resampled_signal = scipy.ndimage.zoom(signal, desired_length / len(signal))
    return resampled_signal


def _resample_fft(signal, desired_length):
    resampled_signal = scipy.signal.resample(signal, desired_length)
    return resampled_signal


def _resample_poly(signal, desired_length):
    resampled_signal = scipy.signal.resample_poly(signal, desired_length, len(signal))
    return resampled_signal


def _resample_pandas(signal, desired_length):
    # Convert to Time Series
    index = pd.date_range("20131212", freq="L", periods=len(signal))
    resampled_signal = pd.Series(signal, index=index)

    # Create resampling factor
    resampling_factor = str(np.round(1 / (desired_length / len(signal)), 6)) + "L"

    # Resample
    resampled_signal = resampled_signal.resample(resampling_factor).bfill().values

    # Sanitize
    resampled_signal = _resample_sanitize(resampled_signal, desired_length)

    return resampled_signal
